package com.hello2morrow.sonargraph.languageprovider.java.controller.system;

import static org.junit.Assert.assertEquals
import static org.junit.Assert.assertFalse
import static org.junit.Assert.assertNotNull
import static org.junit.Assert.assertTrue

import javax.xml.datatype.DatatypeFactory

import org.junit.After
import org.junit.Before
import org.junit.Test

import com.hello2morrow.foundation.activity.DefaultWorkerContext
import com.hello2morrow.foundation.file.FileUtility
import com.hello2morrow.foundation.utilities.OperationResult
import com.hello2morrow.foundation.utilities.OperationResultWithOutcome
import com.hello2morrow.sonargraph.core.command.common.MockUserInterfaceAdapterDevelopment.PluginResourceProviderDevelopment
import com.hello2morrow.sonargraph.core.controller.system.AnalyzerExtension
import com.hello2morrow.sonargraph.core.controller.system.VirtualModelExtension
import com.hello2morrow.sonargraph.core.controllerinterface.system.IIssueExtension
import com.hello2morrow.sonargraph.core.controllerinterface.system.IVirtualModelExtension
import com.hello2morrow.sonargraph.core.model.common.IIssueId
import com.hello2morrow.sonargraph.core.model.element.CoreIssueId
import com.hello2morrow.sonargraph.core.model.element.DependencyIssue
import com.hello2morrow.sonargraph.core.model.element.DependencyWildcardPattern
import com.hello2morrow.sonargraph.core.model.element.ElementWithIssues
import com.hello2morrow.sonargraph.core.model.element.Issue
import com.hello2morrow.sonargraph.core.model.element.IssueType
import com.hello2morrow.sonargraph.core.model.element.NameFilter
import com.hello2morrow.sonargraph.core.model.element.NamedElementWildcardPattern
import com.hello2morrow.sonargraph.core.model.element.Priority
import com.hello2morrow.sonargraph.core.model.path.CoreFileType
import com.hello2morrow.sonargraph.core.model.path.VirtualModels
import com.hello2morrow.sonargraph.core.model.resolution.FixDefinition
import com.hello2morrow.sonargraph.core.model.resolution.IgnoreDefinition
import com.hello2morrow.sonargraph.core.model.resolution.IssueFilter
import com.hello2morrow.sonargraph.core.model.resolution.ResolutionMode
import com.hello2morrow.sonargraph.core.model.resolution.ToDoDefinition
import com.hello2morrow.sonargraph.core.model.system.Files
import com.hello2morrow.sonargraph.core.model.system.Installation
import com.hello2morrow.sonargraph.core.model.system.ModifiableModel
import com.hello2morrow.sonargraph.core.model.system.PrepareRefreshResult
import com.hello2morrow.sonargraph.core.model.system.SoftwareSystem
import com.hello2morrow.sonargraph.core.model.system.VirtualModel
import com.hello2morrow.sonargraph.core.model.system.ISoftwareSystemProvider.OpenOption
import com.hello2morrow.sonargraph.core.model.system.ISoftwareSystemProvider.RefreshOption
import com.hello2morrow.sonargraph.core.model.workspace.Workspace
import com.hello2morrow.sonargraph.core.persistence.virtualmodel.VirtualModelPersistence
import com.hello2morrow.sonargraph.languageprovider.java.foundation.common.JavaLanguage
import com.hello2morrow.sonargraph.languageprovider.java.foundation.common.JavaTestFixture
import com.hello2morrow.sonargraph.languageprovider.java.model.element.JavaIssueId
import com.hello2morrow.sonargraph.languageprovider.java.model.programming.JavaPackageFragment
import com.hello2morrow.sonargraph.languageprovider.java.model.system.JavaModule

import de.schlichtherle.truezip.file.TFile
import de.schlichtherle.truezip.file.TFileInputStream

public final class VirtualModelExtensionPersistenceTest extends AbstractJavaBasedTest {
    private static final String TEST_MODEL_SAVE = "testSaveModel";
    private static final String DESCRIPTION = "description";
    private static final String TEST_MODEL_LOAD = "testLoadModel";
    private static final String PATH_PROJECT_13 = JavaTestFixture.PROJECT_13;
    private static final String MODELS_PATH = PATH_PROJECT_13 + "/Models";
    private static final String PROJECT_BACKUP = "./src/test/project_13/project_13_backup.sonargraph";

    @Before
    public void init() {
        TFile modelsDir = new TFile(MODELS_PATH);
        for (TFile file : modelsDir.listFiles()) {
            if (file.getName().equals("Modifiable.vm") || file.getName().equals("testLoadModel.vm") || file.getName().equals("VirtualModel_To_Be_Ignored.vm_")) {
                continue
            }

            System.out.println "Deleting file: " + file.getAbsolutePath()
            file.rm()
        }

        JavaTestFixture.copyDirectory(new TFile(PATH_PROJECT_13), new TFile(PROJECT_BACKUP));
    }

    //Disable this method, if you need to examine the output file for the virtual model
    @After
    public void tearDown() throws IOException
    {
        TFile backupDir = new TFile(PROJECT_BACKUP);
        JavaTestFixture.copyDirectory(backupDir, new TFile(PATH_PROJECT_13));
        backupDir.rm_r();
    }

    @Test
    public void testSave()
    {
        getController().openSoftwareSystem(DefaultWorkerContext.INSTANCE, new TFile(PATH_PROJECT_13), EnumSet.noneOf(OpenOption.class));
        SoftwareSystem system = getController().getSoftwareSystem();
        getController().refresh(DefaultWorkerContext.INSTANCE, new PrepareRefreshResult(Collections.singleton(JavaLanguage.INSTANCE), EnumSet.of(RefreshOption.REFRESH_PARSER_MODEL), true));

        waitForAnalyzers(system);

        IIssueExtension issueExtension = system.getExtension(IIssueExtension.class);

        OperationResultWithOutcome<ModifiableModel> createModelResult = system.getExtension(VirtualModelExtension.class).createModifiableModel(DefaultWorkerContext.INSTANCE, TEST_MODEL_SAVE, DESCRIPTION, system.getParserModel(),true);
        assertNotNull("Failed to create modifiable model", createModelResult.getOutcome());
        assertTrue(createModelResult.toString(), createModelResult.isSuccess());

        waitForAnalyzers(system);

        ModifiableModel testModel = createModelResult.getOutcome();
        assertTrue("Unexpected type '" + testModel.getClass().getCanonicalName() + "'", testModel instanceof ModifiableModel);
        VirtualModel currentModel = system.getCurrentModel();
        assertTrue(currentModel.isModifiable());
        assertEquals(testModel, currentModel);

        IssueFilter filter = new IssueFilter();
        filter.add(ResolutionMode.NONE);

        List<Issue> issues = currentModel.getIssueList(filter).getIssues();

        //Ignore issue
        Issue issueToIgnore = issues.get(0);
        List<Issue> issuesToIgnore = new ArrayList<>(1);
        IIssueId idOfIssueToIgnore = issueToIgnore.getId();

        issuesToIgnore.add(issueToIgnore);
        String ignoreDescription = "Ignore description";
        issueExtension.ignore(DefaultWorkerContext.INSTANCE, issuesToIgnore, ignoreDescription);

        waitForAnalyzers(system);

        //Fix issue
        Issue issueToFix = issues.get(1);
        IIssueId idOfIssueToFix = issueToFix.getId();
        List<Issue> issuesToFix = new ArrayList<>(1);
        issuesToFix.add(issueToFix);
        String fixDescription = "Fix description";
        String assignedTo = "h2m";
        issueExtension.fix(DefaultWorkerContext.INSTANCE, issuesToFix, fixDescription, assignedTo, Priority.MEDIUM);

        waitForAnalyzers(system);

        //ToDo issue
        List<JavaModule> javaModules = system.getUniqueExistingChild(Workspace.class).getChildrenRecursively(JavaModule.class, JavaPackageFragment.class);
        List<ElementWithIssues> elementsForToDo = new ArrayList<>();
        for (JavaModule module : javaModules)
        {
            elementsForToDo.add(module);
        }
        assertFalse(elementsForToDo.isEmpty());
        String toDoComment = "Examine these elements";
        issueExtension.createToDo(DefaultWorkerContext.INSTANCE, elementsForToDo, toDoComment, assignedTo, Priority.HIGH);

        waitForAnalyzers(system);

        //Dependency issue
        List<Issue> dependencyIssues = currentModel.getIssueList(null, CoreIssueId.DEPENDENCY_TO_EXCLUDED_INTERNAL_COMPONENT).getIssues();
        assertFalse(dependencyIssues.isEmpty());

        DependencyIssue dependencyIssue = dependencyIssues.find
        {
            if (it.getDependencyFromName().equals("com.hello2morrow.foundation.utilities.Platform\$JavaVersion"))
            {
                return true;
            }
            return false;
        }

        String fixDependencyDescription = "Fix dependency";
        issueExtension.fix(DefaultWorkerContext.INSTANCE, Arrays.asList(dependencyIssue), fixDependencyDescription, assignedTo, Priority.HIGH);

        waitForAnalyzers(system);

        ModifiableModel modifiableModel = (ModifiableModel) testModel;
        assertTrue(modifiableModel.needsSave());

        OperationResult result = system.getExtension(VirtualModelExtension.class).saveModifiableModel(testModel);
        assertTrue(result.isSuccess());
        TFile modelFile = new TFile(MODELS_PATH, "testSaveModel.vm");
        assertTrue(modelFile.exists());
        assertFalse(modifiableModel.needsSave());

        def virtualModelXml = new XmlSlurper().parse(new TFileInputStream(modelFile));
        assertEquals('virtualModel', virtualModelXml.name());
        assertEquals('resolutions', virtualModelXml.resolutions.name());
        assertEquals(4, virtualModelXml.resolutions.children().size());

        def ignoreResolution = virtualModelXml.resolutions.ignore[0];
        assertEquals(ignoreDescription, ignoreResolution.'@description'.toString());
        assertEquals(idOfIssueToIgnore.getClass().getCanonicalName(), ignoreResolution.type.'@class'.toString());
        assertEquals(idOfIssueToIgnore.getStandardName(), ignoreResolution.type.'@name'.toString());

        def fixResolution = virtualModelXml.resolutions.fix[0];
        assertEquals(fixDescription, fixResolution.'@description'.toString());
        assertEquals(idOfIssueToFix.getClass().getCanonicalName(), fixResolution.type.'@class'.toString());
        assertEquals(idOfIssueToFix.getStandardName(), fixResolution.type.'@name'.toString());
        assertEquals(assignedTo, fixResolution.'@assignedTo'.toString());
        assertEquals(Priority.MEDIUM.getStandardName(), fixResolution.'@priority'.toString());

        def fixDependencyResolution = virtualModelXml.resolutions.fix[1];
        assertEquals('com.hello2morrow.sonargraph.core.model.element.CoreIssueId', fixDependencyResolution.type.'@class'.toString());
        assertEquals(CoreIssueId.DEPENDENCY_TO_EXCLUDED_INTERNAL_COMPONENT.getStandardName(), fixDependencyResolution.type.'@name'.toString());
        assertEquals(assignedTo, fixDependencyResolution.'@assignedTo'.toString());
        assertEquals(Priority.HIGH.getStandardName(), fixDependencyResolution.'@priority'.toString());
        assertEquals('Workspace:Module_1:./src_1:com:hello2morrow:foundation:utilities:IStandardEnumeration.java:IStandardEnumeration', fixDependencyResolution.dependencyPattern.'@to'.toString());

        def toDoResolution = virtualModelXml.resolutions.todo[0];
        assertEquals(toDoComment, toDoResolution.'@description'.toString());
        assertEquals('com.hello2morrow.sonargraph.core.model.element.CoreIssueId', toDoResolution.type.'@class'.toString());
        assertEquals('Todo', toDoResolution.type.'@name'.toString());
        assertEquals(assignedTo, toDoResolution.'@assignedTo'.toString());
        assertEquals(Priority.HIGH.getStandardName(), toDoResolution.'@priority'.toString());
    }

    //Added extra method to quickly identify the usage within this test case.
    private waitForAnalyzers(SoftwareSystem system)
    {
        system.getExtension(AnalyzerExtension.class).waitForAnalyzersToComplete()
    }

    private void refresh(final Installation installation, final SoftwareSystem softwareSystem, final ModifiableModel model, final OperationResult result)
    {
        assert softwareSystem != null;
        assert model != null : "Parameter 'model' of method 'refresh' must not be null";
        assert result != null : "Parameter 'result' of method 'refresh' must not be null";

        final TFile file = model.getFile();
        if (file.exists())
        {
            model.removeChildren();
            final VirtualModelPersistence persistence = new VirtualModelPersistence(installation.getVersion(), softwareSystem.getClassLoader(), new PluginResourceProviderDevelopment());
            final OperationResult resultLoad = persistence.load(file, model);
            result.addMessagesFrom(resultLoad);
        }
    }

    @Test
    public void testLoad()
    {
        getController().openSoftwareSystem(DefaultWorkerContext.INSTANCE, new TFile(PATH_PROJECT_13), EnumSet.noneOf(OpenOption.class));
        SoftwareSystem system = getController().getSoftwareSystem();
        getController().refresh(DefaultWorkerContext.INSTANCE,
                new PrepareRefreshResult(Collections.singleton(JavaLanguage.INSTANCE), EnumSet.of(RefreshOption.REFRESH_PARSER_MODEL), true));
        waitForAnalyzers(system);

        List<ModifiableModel> available = system.getExtension(IVirtualModelExtension.class).getAvailableModifiableModels();
        assertEquals("Exactly 2 virtual models expected", 2, available.size());
        VirtualModels virtualModels = system.getUniqueExistingChild(Files.class).getVirtualModels();
        List<VirtualModel> models = virtualModels.getChildrenRecursively(new NameFilter(FileUtility.REL_PATH_START + TEST_MODEL_LOAD + CoreFileType.MODEL.getDefaultExtension()), VirtualModel.class);
        assertTrue("Model not found: "+FileUtility.REL_PATH_START + TEST_MODEL_LOAD + CoreFileType.MODEL.getDefaultExtension(), models.size() == 1);
        VirtualModel model = models.get(0);

        OperationResult result = new OperationResult("Refresh model");
        refresh(getController().getInstallation(), system, model, result);
        assertTrue(result.toString(), result.isSuccess());
        DatatypeFactory datatypeFactory = DatatypeFactory.newInstance();

        checkIgnoreResolution(model, datatypeFactory);
        checkFixResolutions(model, datatypeFactory);
        checkToDoResolution(model, datatypeFactory);
    }

    private checkIgnoreResolution(VirtualModel model, DatatypeFactory datatypeFactory) {
        List<IgnoreDefinition> ignoreResolutions = model.getChildren(IgnoreDefinition.class);
        assertEquals(1, ignoreResolutions.size());

        IgnoreDefinition ignoreResolution = ignoreResolutions.get(0);
        assertEquals("Ignore description", ignoreResolution.getDescription());
        Calendar calendar = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        calendar.setTime(ignoreResolution.getDate());
        assertEquals("2013-09-06T11:32:46.083Z", datatypeFactory.newXMLGregorianCalendar(calendar).toString());
        IssueType type = ignoreResolution.getUniqueChild(IssueType.class);
        assertEquals(JavaIssueId.JAVA_FILE_UNPARSEABLE, type.getIssueId());

        List<NamedElementWildcardPattern> patterns = ignoreResolution.getChildren(NamedElementWildcardPattern.class);
        assertEquals(1, patterns.size());
        assertEquals("./com/hello2morrow/foundation/EmptyIterator.java", patterns.get(0).getName());
    }

    private checkFixResolutions(VirtualModel model, DatatypeFactory datatypeFactory) {
        List<FixDefinition> fixResolutions = model.getChildren(FixDefinition.class);
        assertEquals(3, fixResolutions.size());

        FixDefinition fixResolution = fixResolutions.get(0);
        assertEquals("Fix description", fixResolution.getDescription());
        GregorianCalendar calendar = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        calendar.setTime(fixResolution.getDate());
        assertEquals("2013-09-06T11:32:46.099Z", datatypeFactory.newXMLGregorianCalendar(calendar).toString());
        assertEquals("h2m", fixResolution.getAssignee());
        assertEquals(Priority.MEDIUM, fixResolution.getPriority())
        IssueType type = fixResolution.getUniqueChild(IssueType.class);
        assertEquals(JavaIssueId.CLASS_FILE_OUT_OF_DATE, type.getIssueId());

        List<NamedElementWildcardPattern> patterns = fixResolution.getChildren(NamedElementWildcardPattern.class);
        assertEquals(1, patterns.size());
        assertEquals("./com/hello2morrow/foundation/utilities/NoDate.class", patterns.get(0).getName());

        FixDefinition fixDependencyResolution = fixResolutions.get(1);
        assertEquals("Fix dependency", fixDependencyResolution.getDescription());
        calendar = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        calendar.setTime(fixDependencyResolution.getDate());
        assertEquals("2013-09-06T11:32:46.129Z", datatypeFactory.newXMLGregorianCalendar(calendar).toString());
        assertEquals("h2m", fixDependencyResolution.getAssignee());
        assertEquals(Priority.HIGH, fixDependencyResolution.getPriority())
        type = fixDependencyResolution.getUniqueChild(IssueType.class);
        assertEquals(CoreIssueId.DEPENDENCY_TO_EXCLUDED_INTERNAL_COMPONENT, type.getIssueId());

        FixDefinition fixThresholdResolution = fixResolutions.get(2);
        assertEquals("Fix it.", fixThresholdResolution.getDescription());
        calendar = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        calendar.setTime(fixThresholdResolution.getDate());
        assertEquals("2013-10-10T20:11:52.989Z", datatypeFactory.newXMLGregorianCalendar(calendar).toString());
        assertEquals("h2m", fixThresholdResolution.getAssignee());
        assertEquals(Priority.MEDIUM, fixThresholdResolution.getPriority())
        type = fixThresholdResolution.getUniqueChild(IssueType.class);
        assertEquals(CoreIssueId.THRESHOLD_VIOLATION, type.getIssueId());

        List<DependencyWildcardPattern> dependencyPatterns = fixDependencyResolution.getChildren(DependencyWildcardPattern.class);
        assertEquals(1, dependencyPatterns.size());
        DependencyWildcardPattern dependencyPattern = dependencyPatterns.get(0);
        assertEquals('Module_1:com.hello2morrow.foundation.utilities.Platform$JavaVendor', dependencyPattern.getFromPattern());
        assertEquals("Module_1:com.hello2morrow.foundation.utilities.IStandardEnumeration", dependencyPattern.getToPattern());
    }

    private checkToDoResolution(VirtualModel model, DatatypeFactory datatypeFactory) {
        List<ToDoDefinition> toDoResolutions = model.getChildren(ToDoDefinition.class);
        assertEquals(1, toDoResolutions.size());

        ToDoDefinition toDoResolution = toDoResolutions.get(0);
        assertEquals("Examine these elements", toDoResolution.getDescription());
        GregorianCalendar calendar = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
        calendar.setTime(toDoResolution.getDate());
        assertEquals("2013-09-06T11:32:46.106Z", datatypeFactory.newXMLGregorianCalendar(calendar).toString());
        assertEquals("h2m", toDoResolution.getAssignee());
        assertEquals(Priority.HIGH, toDoResolution.getPriority())

        List<NamedElementWildcardPattern> patterns = toDoResolution.getChildren(NamedElementWildcardPattern.class);
        assertEquals(2, patterns.size());
        assertEquals("Module_1", patterns.get(0).getName());
        assertEquals("Module_2", patterns.get(1).getName())
    }
}